Attribute VB_Name = "Module1"
Option Explicit

Private Declare Sub CpyMem Lib "kernel32" _
Alias "RtlMoveMemory" ( _
    pDst As Any, _
    pSrc As Any, _
    ByVal cBytes As Long _
)

Private Declare Function VirtualAlloc Lib "kernel32" ( _
    ByVal lpAddress As Long, _
    ByVal dwSize As Long, _
    ByVal flAllocationType As Long, _
    ByVal flProtect As Long _
) As Long

Private Declare Function VirtualFree Lib "kernel32" ( _
    ByVal lpAddress As Long, _
    ByVal dwSize As Long, _
    ByVal dwFreeType As Long _
) As Long

Private Declare Function CallWindowProc Lib "user32" _
Alias "CallWindowProcA" ( _
    ByVal adr As Long, _
    ByVal p1 As Long, _
    ByVal p2 As Long, _
    ByVal p3 As Long, _
    ByVal p4 As Long _
) As Long

Private Enum VirtualFreeTypes
    MEM_DECOMMIT = &H4000
    MEM_RELEASE = &H8000
End Enum

Private Enum VirtualAllocTypes
    MEM_COMMIT = &H1000
    MEM_RESERVE = &H2000
    MEM_RESET = &H8000
    MEM_LARGE_PAGES = &H20000000
    MEM_PHYSICAL = &H100000
    MEM_WRITE_WATCH = &H200000
End Enum

Private Enum VirtualAllocPageFlags
    PAGE_EXECUTE = &H10
    PAGE_EXECUTE_READ = &H20
    PAGE_EXECUTE_READWRITE = &H40
    PAGE_EXECUTE_WRITECOPY = &H80
    PAGE_NOACCESS = &H1
    PAGE_READONLY = &H2
    PAGE_READWRITE = &H4
    PAGE_WRITECOPY = &H8
    PAGE_GUARD = &H100
    PAGE_NOCACHE = &H200
    PAGE_WRITECOMBINE = &H400
End Enum

Private Type COM_IUnknown
    QueryInterface  As Long
    AddRef          As Long
    Release         As Long
End Type

Private Type COM_IRecordInfo
    iunk            As COM_IUnknown
    RecordInit      As Long
    RecordClear     As Long
    RecordCopy      As Long
    GetGuid         As Long
    GetName         As Long
    GetSize         As Long
    GetTypeInfo     As Long
    GetField        As Long
    GetFieldNoCopy  As Long
    PutField        As Long
    PutFieldNoCopy  As Long
    GetFieldNames   As Long
    IsMatchingType  As Long
    RecordCreate    As Long
    RecordCreateCpy As Long
    RecordDestroy   As Long
End Type

Private Type VARIANTARG
    vt              As Integer

    wRsvd1          As Integer
    wRsvd2          As Integer
    wRsvd3          As Integer

    lngValHi        As Long
    lngValLo        As Long
End Type

Public Type UDT_Info
    name            As String
    size            As Long
    fields          As Long
    fieldnames()    As String
    fieldvalues()   As Variant
End Type

Public Function UDT_Get_Info( _
    udt As Variant _
) As UDT_Info

    Dim udtVariant  As VARIANTARG
    Dim ppRecInfo   As Long
    Dim pRecInfo    As Long
    Dim lngSize     As Long
    Dim lngFields   As Long
    Dim i           As Long
    Dim strName     As String
    Dim strNames()  As String
    Dim varValues() As Variant
    Dim oRecInfo    As COM_IRecordInfo

    ' make the part of the variant important to us accessible
    CpyMem udtVariant, udt, Len(udtVariant)

    ' must be an UDT
    If udtVariant.vt <> 36 Then Exit Function

    ' lngValLo is a pointer to a pointer to the IRecordInfo VTable
    ' lngValHi is a pointer to the UDT

    ' copy the VTable
    ppRecInfo = udtVariant.lngValLo
    CpyMem pRecInfo, ByVal ppRecInfo, 4
    CpyMem oRecInfo, ByVal pRecInfo, Len(oRecInfo)

    ' increment reference counter
    CallPointer oRecInfo.iunk.AddRef, ppRecInfo

    ' stuff about the UDT
    CallPointer oRecInfo.GetName, ppRecInfo, VarPtr(strName)
    CallPointer oRecInfo.GetSize, ppRecInfo, VarPtr(lngSize)
    CallPointer oRecInfo.GetFieldNames, ppRecInfo, VarPtr(lngFields), 0

    If lngFields > 0 Then
        ReDim strNames(lngFields - 1) As String
        ReDim varValues(lngFields - 1) As Variant

        ' get field names
        CallPointer oRecInfo.GetFieldNames, ppRecInfo, _
                                            VarPtr(lngFields), _
                                            VarPtr(strNames(0))

        ' get field values
        For i = 0 To lngFields - 1
            CallPointer oRecInfo.GetField, ppRecInfo, _
                                           udtVariant.lngValHi, _
                                           StrPtr(strNames(i)), _
                                           VarPtr(varValues(i))
        Next
    End If

    ' decrement reference counter
    CallPointer oRecInfo.iunk.Release, ppRecInfo

    With UDT_Get_Info
        .name = strName
        .size = lngSize
        .fields = lngFields
        .fieldnames = strNames
        .fieldvalues = varValues
    End With
End Function

Private Function CallPointer( _
    ByVal fnc As Long, _
    ParamArray params() _
) As Long

    Dim pMemory             As Long
    Dim cBytesMemory        As Long
    Dim pASM                As Long
    Dim i                   As Integer
    Dim j                   As Integer

    If fnc = 0 Then
        Err.Raise "CallPointer called with fnc=0!"
    End If

    cBytesMemory = &HEC00&
    pMemory = VirtualAlloc(0, cBytesMemory, MEM_COMMIT, PAGE_EXECUTE_READWRITE)
    If pMemory = 0 Then Exit Function

    pASM = pMemory

    AddByte pASM, &H58                  ' POP EAX
    AddByte pASM, &H59                  ' POP ECX
    AddByte pASM, &H59                  ' POP ECX
    AddByte pASM, &H59                  ' POP ECX
    AddByte pASM, &H59                  ' POP ECX
    AddByte pASM, &H50                  ' PUSH EAX

    For i = UBound(params) To 0 Step -1
        If IsArray(params(i)) Then
            For j = UBound(params(i)) To 0 Step -1
                AddPush pASM, CLng(params(i)(j))    ' PUSH dword
            Next
        Else
            AddPush pASM, CLng(params(i))           ' PUSH dword
        End If
    Next

    AddCall pASM, fnc                   ' CALL rel addr
    AddByte pASM, &HC3                  ' RET

    CallPointer = CallWindowProc(pMemory, 0, 0, 0, 0)

    VirtualFree pMemory, cBytesMemory, MEM_DECOMMIT
End Function

Private Sub AddPush(pASM As Long, lng As Long)
    AddByte pASM, &H68
    AddLong pASM, lng
End Sub

Private Sub AddCall(pASM As Long, addr As Long)
    AddByte pASM, &HE8
    AddLong pASM, addr - pASM - 4
End Sub

Private Sub AddLong(pASM As Long, lng As Long)
    CpyMem ByVal pASM, lng, 4
    pASM = pASM + 4
End Sub

Private Sub AddByte(pASM As Long, bt As Byte)
    CpyMem ByVal pASM, bt, 1
    pASM = pASM + 1
End Sub
